/*
 * mlx90614.c
 *
 *  Created on: Apr 27, 2025
 *      Author: Alex, www.aeq-web.com
 */

#include <mlx90614.h>
#include "i2c.h"
#include "main.h"
#include "hw_gpio.h"

static uint8_t mlx90614_address = 0x5A;
extern I2C_HandleTypeDef hi2c2;

void mlx90614_init(uint8_t addr) {

	GPIO_InitTypeDef pinConfig = { 0 };
	pinConfig.Mode = GPIO_MODE_OUTPUT_PP;
	pinConfig.Pull = GPIO_NOPULL;
	pinConfig.Speed = GPIO_SPEED_FREQ_VERY_HIGH;

	HW_GPIO_Init( GPIOC, GPIO_PIN_13, &pinConfig);
	HW_GPIO_Write( GPIOC, GPIO_PIN_13, 1);

	HAL_Delay(2);

	mlx90614_address = addr;

	MX_I2C2_Init(I2C_CLOCK_1MHz);

	HAL_I2C_StateTypeDef value;
	do {
		value = HAL_I2C_GetState(&hi2c2);
	} while (value != HAL_I2C_STATE_READY);

	HAL_Delay(100);

}

void mlx90614_deinit(void) {
	i2c_mod2_deinit();
	HW_GPIO_Deinit( GPIOC, GPIO_PIN_13, 0);
}

bool mlx90614_connected(void) {
	return (HAL_I2C_IsDeviceReady(&hi2c2, mlx90614_address << 1, 3, 10)
			== HAL_OK);
}

float mlx90614_read_surface_temp() {
	uint8_t data[3] = { 0 };
	HAL_Delay(100);
	if (i2c_mod2_read_register(mlx90614_address << 1, 0x07, 1, data, 3)
			!= HAL_OK) {
		return -999.0f;
	}

	uint16_t raw = ((uint16_t) data[1] << 8) | data[0];
	float temp_c = raw * 0.02f - 273.15f;
	return temp_c;
}

float mlx90614_read_surface_temp_avg(uint8_t samples) {
	if (samples == 0)
		return -999.0f;
	HAL_Delay(100);
	float sum = 0.0f;
	uint8_t valid = 0;

	for (uint8_t i = 0; i < samples; i++) {
		uint8_t data[3] = { 0 };

		if (i2c_mod2_read_register(mlx90614_address << 1, 0x07, 1, data, 3)
				!= HAL_OK) {
			HAL_Delay(100);  // wait before next try
			continue;
		}

		// Check error flag in MSB
		if (data[1] & 0x80) {
			HAL_Delay(100);
			continue;
		}

		uint16_t raw = ((uint16_t) data[1] << 8) | data[0];
		float temp_c = raw * 0.02f - 273.15f;

		sum += temp_c;
		valid++;

		HAL_Delay(100);  // Optional small delay between samples
	}

	return (valid > 0) ? (sum / valid) : -999.0f;
}

float mlx90614_read_internal_temp() {
	uint8_t data[3] = { 0 };
	HAL_Delay(100);
	if (i2c_mod2_read_register(mlx90614_address << 1, 0x06, 1, data, 3)
			!= HAL_OK) {
		return -999.0f;
	}

	uint16_t raw = ((uint16_t) data[1] << 8) | data[0];
	float temp_c = raw * 0.02f - 273.15f;
	return temp_c;
}

float mlx90614_read_emissivity() {
	uint8_t data[3];
	HAL_Delay(100);
	if (i2c_mod2_read_register(mlx90614_address << 1, 0x24, 1, data, 3)
			!= HAL_OK) {
		return -1.0f;
	}
	uint16_t raw = ((uint16_t) data[1] << 8) | data[0];
	return (float) raw / 65535.0f;
}

// Calculate CRC8 for SMBus (Polynomial 0x07, Init 0x00)
static uint8_t mlx90614_calculate_pec(uint8_t *data, uint8_t len) {
	uint8_t crc = 0;
	for (uint8_t i = 0; i < len; i++) {
		crc ^= data[i];
		for (uint8_t j = 0; j < 8; j++) {
			if (crc & 0x80)
				crc = (crc << 1) ^ 0x07;
			else
				crc <<= 1;
		}
	}
	return crc;
}

bool mlx90614_set_emissivity(float emissivity) {
	if (emissivity < 0.1f || emissivity > 1.0f)
		return false;

	uint16_t value = (uint16_t) (emissivity * 65535.0f);
	uint8_t eeprom_reg = 0x24;

	// Step 1: delete EEPROM (Write 0x0000)
	uint8_t erase_buf[4] = { eeprom_reg, 0x00, 0x00, 0x00 };
	erase_buf[3] = mlx90614_calculate_pec((uint8_t[] ) { mlx90614_address << 1,
					erase_buf[0], erase_buf[1], erase_buf[2] }, 4);
	HAL_Delay(100);
	if (HAL_I2C_Master_Transmit(&hi2c2, mlx90614_address << 1, erase_buf, 4,
			100) != HAL_OK)
		return false;
	HAL_Delay(10);

	// Step 2: Write new value to EEPROM
	uint8_t data_lsb = value & 0xFF;
	uint8_t data_msb = value >> 8;
	uint8_t pec = mlx90614_calculate_pec((uint8_t[] ) { mlx90614_address << 1,
					eeprom_reg, data_lsb, data_msb }, 4);

	uint8_t write_buf[4] = { eeprom_reg, data_lsb, data_msb, pec };

	if (HAL_I2C_Master_Transmit(&hi2c2, mlx90614_address << 1, write_buf, 4,
			100) != HAL_OK)
		return false;
	HAL_Delay(10);

	// Step 3: Read & Verify
	uint8_t read_buf[3] = { 0 };
	if (i2c_mod2_read_register(mlx90614_address << 1, eeprom_reg, 1, read_buf,
			3) != HAL_OK)
		return false;

	uint16_t read_val = ((uint16_t) read_buf[1] << 8) | read_buf[0];
	return (read_val == value);
}
